@@ -1,8 +1,8 @@
 import traceback
 import threading
-import select
 import socket
 import struct
+import errno
 import time
 import os
 import io
@@ -32,9 +32,9 @@ from pyroute2.netlink import IPRCMD_DISCOVER
 from pyroute2.netlink.generic import mgmtmsg
 from pyroute2.netlink.generic import envmsg
 from pyroute2.iocore import NLT_CONTROL
-from pyroute2.iocore import NLT_NOOP
 from pyroute2.iocore import NLT_RESPONSE
 from pyroute2.iocore import NLT_DGRAM
+from pyroute2.iocore.loop import IOLoop
 from pyroute2.iocore.addrpool import AddrPool
 
 try:
@@ -42,24 +42,9 @@ try:
 except ImportError:
     import urllib.parse as urlparse
 
-try:
-    import Queue
-except ImportError:
-    import queue as Queue
-_QUEUE_MAXSIZE = 4096
-
 
 C_ADDR_START = 3
 
-##
-# FIXME: achtung, monkeypatch!
-#
-# Android QPython 2.7 platform has no AF_UNIX, so add some
-# invalid value just not to fail on checks.
-#
-if not hasattr(socket, 'AF_UNIX'):
-    socket.AF_UNIX = 65535
-
 
 def _monkey_handshake(self):
     ##
@@ -264,22 +249,19 @@ class MarshalEnv(Marshal):
     msg_map = {NLMSG_TRANSPORT: envmsg}
 
 
-class IOBroker(threading.Thread):
+class IOBroker(object):
     def __init__(self,
                  addr=0x01000000,
-                 broadcast=0xffffffff):
-        threading.Thread.__init__(self, name='Netlink I/O core')
-        self.setDaemon(True)
+                 broadcast=0xffffffff,
+                 ioloop=None):
         self.pid = os.getpid()
-        self._run_event = threading.Event()
-        self._sctl_event = threading.Event()
         self._stop_event = threading.Event()
         self._reload_event = threading.Event()
         self.addr = addr
         self.broadcast = broadcast
         self.marshal = MarshalEnv()
-        self.ports = AddrPool()
-        self.nonces = AddrPool()
+        self.ports = AddrPool(minaddr=0xff)
+        self.nonces = AddrPool(minaddr=0xfff)
         self.active_sys = {}
         self.local = {}
         self.links = {}
@@ -291,31 +273,42 @@ class IOBroker(threading.Thread):
         self._xlist = set()
         # routing
         self.masquerade = {}      # {int: MasqRecord()...}
-        self.recv_methods = {}    # {socket: recv_method, ...}
         self.clients = set()      # set(socket, socket...)
         self.servers = set()      # set(socket, socket...)
         self.controls = set()     # set(socket, socket...)
         self.subscribe = {}
         self.providers = {}
-        self.queue = Queue.Queue(_QUEUE_MAXSIZE)
         self._cid = list(range(1024))
         # secret; write non-zero byte as terminator
         self.secret = os.urandom(15)
         self.secret += b'\xff'
         self.uuid = uuid.uuid4()
-        # control in-process communication only
-        self.sctl, self.control = pairPipeSockets()
-        self.add_client(self.sctl)
-        self.controls.add(self.sctl)
         # masquerade cache expiration
         self._expire_thread = threading.Thread(target=self._expire_masq,
                                                name='Masquerade cache')
         self._expire_thread.setDaemon(True)
-        self._expire_thread.start()
-        self._dequeue_thread = threading.Thread(target=self._dequeue,
-                                                name='Buffer queue')
-        self._dequeue_thread.setDaemon(True)
-        self._dequeue_thread.start()
+        self.ioloop = ioloop or IOLoop()
+
+    def handle_connect(self, fd, event):
+        (client, addr) = fd.accept()
+        self.add_client(client)
+        # announce address
+        # .. _ioc-connect:
+        rsp = mgmtmsg()
+        rsp['header']['type'] = NLMSG_CONTROL
+        rsp['cmd'] = IPRCMD_ACK
+        rsp['attrs'] = [['IPR_ATTR_ADDR', self.addr]]
+        rsp.encode()
+        ne = envmsg()
+        ne['dst'] = self.broadcast
+        ne['header']['pid'] = os.getpid()
+        ne['header']['type'] = NLMSG_TRANSPORT
+        ne['header']['flags'] = NLT_CONTROL | NLT_RESPONSE
+        ne['attrs'] = [['IPR_ATTR_CDATA',
+                        rsp.buf.getvalue()]]
+        ne.encode()
+        client.send(ne.buf.getvalue())
+        self.ioloop.register(client, self.route, defer=True)
 
     def alloc_addr(self):
         return self.ports.alloc()
@@ -323,16 +316,6 @@ class IOBroker(threading.Thread):
     def dealloc_addr(self, addr):
         self.ports.free(addr)
 
-    def _dequeue(self):
-        '''
-        Background thread that serves the buffer
-        '''
-        while not self._stop_event.is_set():
-            try:
-                self.route(*self.queue.get())
-            except:
-                pass
-
     def _expire_masq(self):
         '''
         Background thread that expires masquerade cache entries
@@ -381,14 +364,11 @@ class IOBroker(threading.Thread):
                 ne.encode()
                 sock.send(ne.buf.getvalue())
                 # Stop iothread -- shutdown sequence
-                self._stop_event.set()
-                self.queue.put(None)
-                self.control.send(struct.pack('I', 4))
-                return
+                return self.shutdown()
 
             elif cmd['cmd'] == IPRCMD_RELOAD:
                 # Reload io cycle
-                self._reload_event.set()
+                self.ioloop.reload()
                 rsp['cmd'] = IPRCMD_ACK
 
             elif cmd['cmd'] == IPRCMD_SERVE:
@@ -400,12 +380,19 @@ class IOBroker(threading.Thread):
                                                key=key,
                                                cert=cert,
                                                ca=ca)
+                new_sock.setsockopt(socket.SOL_SOCKET,
+                                    socket.SO_REUSEADDR, 1)
                 new_sock.bind(addr)
                 if new_sock.type == socket.SOCK_STREAM:
                     new_sock.listen(16)
-                    self.servers.add(new_sock)
+                    self.ioloop.register(new_sock,
+                                         self.handle_connect)
+                else:
+                    self.ioloop.register(new_sock,
+                                         self.route,
+                                         defer=True)
+                self.servers.add(new_sock)
                 self._rlist.add(new_sock)
-                self.noop()
                 rsp['cmd'] = IPRCMD_ACK
 
             elif cmd['cmd'] == IPRCMD_SHUTDOWN:
@@ -414,18 +401,18 @@ class IOBroker(threading.Thread):
                     if _repr_sockets([old_sock], 'local') == [url]:
                         self._rlist.remove(old_sock)
                         self.servers.remove(old_sock)
-                        self.noop()
+                        self.ioloop.unregister(old_sock)
                         rsp['cmd'] = IPRCMD_ACK
 
             elif cmd['cmd'] == IPRCMD_DISCONNECT:
                 # drop a connection, identified by an addr
                 try:
                     uid = cmd.get_attr('IPR_ATTR_UUID')
-                    self.deregister_link(uid)
+                    old_sock = self.deregister_link(uid)
                     rsp['cmd'] = IPRCMD_ACK
-                    self.noop()
                 except Exception as e:
-                    rsp['attrs'] = [['IPR_ATTR_ERROR', str(e)]]
+                    rsp['attrs'] = [['IPR_ATTR_ERROR',
+                                     traceback.format_exc()]]
 
             elif cmd['cmd'] == IPRCMD_PROVIDE:
                 url = cmd.get_attr('IPR_ATTR_HOST')
@@ -453,6 +440,8 @@ class IOBroker(threading.Thread):
                     established = False
                     uid = str(uuid.uuid4())
 
+                    route = self.route
+
                     if url in self.providers:
                         new_sock = self.providers[url]
                         established = True
@@ -460,11 +449,13 @@ class IOBroker(threading.Thread):
                             new_sock.send(self.gate_local(d, s))
 
                     elif target.scheme == 'netlink':
-                        new_sock = NetlinkSocket(int(target.hostname))
-                        new_sock.bind(int(target.port))
+                        res = target.path.split("/")
+                        new_sock = NetlinkSocket(int(res[1]))
+                        new_sock.bind(int(res[2]))
                         gate = lambda d, s:\
                             new_sock.sendto(self.gate_untag(d, s),
                                             (0, 0))
+                        route = self.route_netlink
 
                     elif target.scheme == 'udp':
                         (new_sock, addr) = _get_socket(url,
@@ -505,11 +496,16 @@ class IOBroker(threading.Thread):
                                               established=established,
                                               remote=remote)
                     link.gate = gate
-                    self.discover[url] = port
+                    self.discover[target.path] = port
                     rsp['attrs'].append(['IPR_ATTR_UUID', uid])
                     rsp['attrs'].append(['IPR_ATTR_ADDR', peer])
+                    try:
+                        self.ioloop.register(new_sock, route,
+                                             defer=True)
+                    except Exception as e:
+                        if e.errno != errno.EEXIST:
+                            raise e
                     rsp['cmd'] = IPRCMD_ACK
-                    self.noop()
 
                 except Exception:
                     rsp['attrs'].append(['IPR_ATTR_ERROR',
@@ -532,6 +528,7 @@ class IOBroker(threading.Thread):
                                          traceback.format_exc()])
 
             elif cmd['cmd'] == IPRCMD_DISCOVER:
+                # .. _ioc-discover:
                 url = cmd.get_attr('IPR_ATTR_HOST')
                 if url in self.discover:
                     rsp['attrs'].append(['IPR_ATTR_ADDR', self.discover[url]])
@@ -557,8 +554,9 @@ class IOBroker(threading.Thread):
         ne['header']['pid'] = pid
         ne['header']['type'] = NLMSG_TRANSPORT
         ne['header']['flags'] = NLT_CONTROL | NLT_RESPONSE
-        ne['dst'] = src
+        # ne['dst'] = src
         ne['src'] = dst
+        ne['ttl'] = 16
         ne['dport'] = sport
         ne['sport'] = dport
         ne['attrs'] = [['IPR_ATTR_CDATA', rsp.buf.getvalue()]]
@@ -567,19 +565,31 @@ class IOBroker(threading.Thread):
 
     def route_forward(self, sock, envelope):
         nonce = envelope['header']['sequence_number']
-        if nonce in self.masquerade:
-            self.unmasq(nonce, envelope)
 
-        # nothing special, just broadcast packet
         envelope['ttl'] -= 1
-        if envelope['ttl'] > 0:
+        if envelope['ttl'] <= 0:
+            return
+
+        if (envelope['dst'] == 0) and (nonce in self.masquerade):
+            return self.unmasq(nonce, envelope)
+        else:
             flags = envelope['header']['flags']
             for (uid, link) in self.remote.items():
                 # by default, send packets only via SOCK_STREAM,
                 # and use SOCK_DGRAM only upon request
-                if ((link.sock.type == socket.SOCK_STREAM) or
-                        (link.sock.type & flags & NLT_DGRAM)):
-                    self.remote[uid].gate(envelope, sock)
+
+                # skip STREAM sockets if NLT_DGRAM is requested
+                if ((link.sock.type == socket.SOCK_STREAM) and
+                        (flags & NLT_DGRAM)):
+                    continue
+
+                # skip DGRAM sockets if NLT_DGRAM is not requested
+                if ((link.sock.type == socket.SOCK_DGRAM) and
+                        not (flags & NLT_DGRAM)):
+                    continue
+
+                # in any other case -- send packet
+                self.remote[uid].gate(envelope, sock)
 
     def unmasq(self, nonce, envelope):
         target = self.masquerade[nonce]
@@ -593,6 +603,7 @@ class IOBroker(threading.Thread):
 
     def route_data(self, sock, envelope):
         nonce = envelope['header']['sequence_number']
+
         if envelope['dport'] in self.local:
             try:
                 self.local[envelope['dport']].gate(envelope, sock)
@@ -623,7 +634,9 @@ class IOBroker(threading.Thread):
         envelope.encode()
         u32['socket'].send(envelope.buf.getvalue())
 
-    def route_netlink(self, sock, data):
+    def route_netlink(self, sock, raw):
+        data = io.BytesIO()
+        data.length = data.write(raw)
         data.seek(8)
         seq = struct.unpack('I', data.read(4))[0]
 
@@ -654,8 +667,9 @@ class IOBroker(threading.Thread):
             envelope['header']['pid'] = \
                 target.envelope['header']['pid']
             envelope['header']['type'] = NLMSG_TRANSPORT
-            envelope['dst'] = target.envelope['src']
+            # envelope['dst'] = target.envelope['src']
             envelope['src'] = target.envelope['dst']
+            envelope['ttl'] = 16
             envelope['dport'] = target.envelope['sport']
             envelope['sport'] = target.envelope['dport']
             envelope['attrs'] = [['IPR_ATTR_CDATA',
@@ -664,13 +678,19 @@ class IOBroker(threading.Thread):
             # target
             target.socket.send(envelope.buf.getvalue())
 
-    def route(self, sock, data):
+    def route(self, sock, raw):
         """
         Route message
         """
-        if isinstance(sock, NetlinkSocket):
-            # netlink packets from the local system
-            return self.route_netlink(sock, data)
+        data = io.BytesIO()
+        data.length = data.write(raw)
+
+        if data.length == 0 and self.ioloop.unregister(sock):
+            if sock in self.clients:
+                self.remove_client(sock)
+            else:
+                self.deregister_link(fd=sock)
+            return
 
         for envelope in self.marshal.parse(data):
             if envelope['dst'] != self.addr:
@@ -680,9 +700,6 @@ class IOBroker(threading.Thread):
             else:
                 # INPUT
                 # a packet for a local system
-                if envelope['header']['flags'] & NLT_NOOP:
-                    # noop packets (drop)
-                    continue
                 if ((envelope['header']['flags'] & NLT_CONTROL) and not
                         (envelope['header']['flags'] & NLT_RESPONSE)):
                     # control packets
@@ -691,11 +708,6 @@ class IOBroker(threading.Thread):
                     # transport packets
                     self.route_data(sock, envelope)
 
-    def recv(self, fd, buf):
-        ret = 0
-        ret += buf.write(fd.recv(16384))
-        return ret, {}
-
     def gate_local(self, envelope, sock):
         # 2. register way back
         nonce = self.nonces.alloc()
@@ -743,54 +755,6 @@ class IOBroker(threading.Thread):
         cmd.decode()
         return cmd
 
-    def noop(self):
-        msg = envmsg()
-        msg['dst'] = self.addr
-        msg['header']['type'] = NLMSG_TRANSPORT
-        msg['header']['flags'] = NLT_NOOP
-        msg.encode()
-        self.control.send(msg.buf.getvalue())
-
-    def command(self, cmd, attrs=[]):
-        msg = mgmtmsg(io.BytesIO())
-        msg['header']['type'] = NLMSG_CONTROL
-        msg['cmd'] = cmd
-        msg['attrs'] = attrs
-        msg.encode()
-        envelope = envmsg()
-        envelope['dst'] = self.addr
-        envelope['header']['type'] = NLMSG_TRANSPORT
-        envelope['header']['flags'] = NLT_CONTROL
-        envelope['attrs'] = [['IPR_ATTR_CDATA', msg.buf.getvalue()]]
-        envelope.encode()
-        self.control.send(envelope.buf.getvalue())
-        envelope = envmsg(self.control.recv())
-        envelope.decode()
-        data = io.BytesIO(envelope.get_attr('IPR_ATTR_CDATA'))
-        rsp = mgmtmsg(data)
-        rsp.decode()
-        return rsp
-
-    def stop(self):
-        try:
-            self.command(IPRCMD_STOP)
-        except OSError:
-            pass
-        self.join()
-
-    def reload(self):
-        '''
-        Reload I/O cycle. Warning: this method should be never
-        called from IOThread, as it will not return.
-        '''
-        self._reload_event.clear()
-        ret = self.command(IPRCMD_RELOAD)
-        # wait max 3 seconds for reload
-        # FIXME: timeout should be configurable
-        self._reload_event.wait(3)
-        assert self._reload_event.is_set()
-        return ret
-
     def register_link(self, uid, port, sock,
                       established=False, remote=False):
         if not established:
@@ -821,6 +785,7 @@ class IOBroker(threading.Thread):
             del self.remote[link.uid]
         else:
             del self.local[link.port]
+        return link.sock
 
     def add_client(self, sock):
         '''
@@ -838,87 +803,13 @@ class IOBroker(threading.Thread):
         self.clients.remove(sock)
         return sock
 
-    def run(self):
-        self._run_event.set()
-        while not self._stop_event.is_set():
-            try:
-                [rlist, wlist, xlist] = select.select(self._rlist, [], [])
-            except:
-                # FIXME: log exceptions
-                traceback.print_exc()
-                continue
-            for fd in rlist:
-                ##
-                #
-                # Incoming remote connections
-                #
-                if fd in self.servers:
-                    try:
-                        (client, addr) = fd.accept()
-                        self.add_client(client)
-                        # announce address
-                        rsp = mgmtmsg()
-                        rsp['header']['type'] = NLMSG_CONTROL
-                        rsp['cmd'] = IPRCMD_ACK
-                        rsp['attrs'] = [['IPR_ATTR_ADDR', self.addr]]
-                        rsp.encode()
-                        ne = envmsg()
-                        ne['dst'] = self.broadcast
-                        ne['header']['pid'] = os.getpid()
-                        ne['header']['type'] = NLMSG_TRANSPORT
-                        ne['header']['flags'] = NLT_CONTROL | NLT_RESPONSE
-                        ne['attrs'] = [['IPR_ATTR_CDATA',
-                                        rsp.buf.getvalue()]]
-                        ne.encode()
-                        client.send(ne.buf.getvalue())
-
-                    except ssl.SSLError:
-                        # FIXME log SSL errors
-                        pass
-                    except:
-                        traceback.print_exc()
-                    continue
-
-                ##
-                #
-                # Receive data from already open connection
-                #
-                # FIXME max length
-                data = io.BytesIO()
-                try:
-                    # recv method
-                    recv = self.recv_methods.get(fd, self.recv)
-                    # fill the routing info and get the data
-                    data.length, rinfo = recv(fd, data)
-                except:
-                    traceback.print_exc()
-                    continue
-
-                ##
-                #
-                # Close socket
-                #
-                if data.length == 0:
-                    if fd in self.clients:
-                        self.remove_client(fd)
-                    else:
-                        self.deregister_link(fd=fd)
-                    continue
-
-                ##
-                #
-                # Route the data
-                #
-                try:
-                    self.queue.put((fd, data))
-                except Exception:
-                    # FIXME: silently drop all exceptions yet
-                    pass
+    def start(self):
+        self._expire_thread.start()
+        self.ioloop.start()
 
+    def shutdown(self):
+        self._stop_event.set()
+        for sock in self.servers:
+            sock.close()
         # shutdown sequence
-        self._rlist.remove(self.sctl)
-        self._wlist.remove(self.sctl)
-        self.sctl.close()
-        self.control.close()
         self._expire_thread.join()
-        self._dequeue_thread.join()
