@@ -1,10 +1,10 @@
-import threading
-import select
 import struct
+import copy
 import os
 import io
 
 from pyroute2.common import uuid32
+from pyroute2.common import debug
 from pyroute2.netlink import NLMSG_CONTROL
 from pyroute2.netlink import NLMSG_TRANSPORT
 from pyroute2.netlink import IPRCMD_ACK
@@ -20,14 +20,15 @@ from pyroute2.netlink import IPRCMD_REMOVE
 from pyroute2.netlink import IPRCMD_DISCOVER
 from pyroute2.netlink import NLMSG_DONE
 from pyroute2.netlink import NLM_F_MULTI
-from pyroute2.netlink import NLM_F_DUMP
-from pyroute2.netlink import NLM_F_REQUEST
 from pyroute2.netlink.generic import mgmtmsg
 from pyroute2.netlink.generic import envmsg
 from pyroute2.iocore import NLT_CONTROL
 from pyroute2.iocore import NLT_RESPONSE
+from pyroute2.iocore import NLT_EXCEPTION
+from pyroute2.iocore.loop import IOLoop
 from pyroute2.iocore.broker import pairPipeSockets
 from pyroute2.iocore.broker import IOBroker
+from pyroute2.iocore.addrpool import AddrPool
 
 try:
     import Queue
@@ -35,22 +36,23 @@ except ImportError:
     import queue as Queue
 _QUEUE_MAXSIZE = 4096
 
+try:
+    import urlparse
+except ImportError:
+    import urllib.parse as urlparse
 
-class IOCore(threading.Thread):
 
-    family = 0
-    groups = 0
+class IOCore(object):
+
     marshal = None
-    do_connect = False
     name = 'Core API'
+    default_target = None
 
-    def __init__(self, debug=False, timeout=3, do_connect=None,
+    def __init__(self, debug=False, timeout=3, do_connect=False,
                  host=None, key=None, cert=None, ca=None,
                  addr=None):
-        threading.Thread.__init__(self, name=self.name)
         addr = addr or uuid32()
         self._timeout = timeout
-        self.iothread = IOBroker(addr=addr)
         self.default_broker = addr
         self.default_dport = 0
         self.uids = set()
@@ -58,154 +60,151 @@ class IOCore(threading.Thread):
         self.callbacks = []     # [(predicate, callback, args), ...]
         self.debug = debug
         self.cid = None
-        self._nonce = 0
-        self._nonce_lock = threading.Lock()
+        self.nonce = AddrPool()
+        self.save = None
         if self.marshal is not None:
             self.marshal.debug = debug
             self.marshal = self.marshal()
         self.buffers = Queue.Queue()
         self._mirror = False
-        self.default_target = 'netlink://%i:%i' % (self.family, self.groups)
-        self.host = host or self.default_target
-        self._run_event = threading.Event()
-        self._stop_event = threading.Event()
-        self._feed_thread = threading.Thread(target=self._feed_buffers,
-                                             name='Reasm and parsing')
-        self._feed_thread.setDaemon(True)
-        self._feed_thread.start()
-        self.setDaemon(True)
-        self.start()
-        self._run_event.wait()
-        if do_connect is not None:
-            self.do_connect = do_connect
-        if self.do_connect:
+        self.host = host
+
+        self.ioloop = IOLoop()
+
+        self.iobroker = IOBroker(addr=addr, ioloop=self.ioloop)
+        self.iobroker.start()
+        self._brs, self.bridge = pairPipeSockets()
+        self.iobroker.add_client(self._brs)
+        self.iobroker.controls.add(self._brs)
+        self.ioloop.register(self._brs,
+                             self.iobroker.route,
+                             defer=True)
+        self.ioloop.register(self.bridge,
+                             self._route,
+                             defer=True)
+        if do_connect:
+            path = urlparse.urlparse(host).path
             (self.default_link,
-             self.default_peer) = self.connect(self.host, key, cert, ca)
-            self.default_dport = self.discover(self.default_target,
+             self.default_peer) = self.connect(self.host,
+                                               key, cert, ca)
+            self.default_dport = self.discover(self.default_target or path,
                                                self.default_peer)
 
-    def run(self):
-        # 1. run iothread
-        self.iothread.start()
-        # 2. connect to iothread
-        self._brs, self.bridge = pairPipeSockets()
-        self.iothread.add_client(self._brs)
-        self.iothread.controls.add(self._brs)
-        self.iothread.reload()
-        self._run_event.set()
-
-        # 3. start to monitor it
-        while not self._stop_event.is_set():
-            try:
-                [rlist, wlist, xlist] = select.select([self.bridge], [], [])
-            except:
-                continue
-            for fd in rlist:
-                data = io.BytesIO()
-                try:
-                    data = fd.recv()
-                except:
-                    continue
-
-                # put data in the queue
-                self.buffers.put(data)
-
-    def _feed_buffers(self):
-        '''
-        Background thread to feed reassembled buffers to the parser
-        '''
-        save = None
-        while True:
-            buf = io.BytesIO()
-            buf.length = buf.write(self.buffers.get())
-            if self._stop_event.is_set():
-                return
+    @debug
+    def _route(self, sock, raw):
+        buf = io.BytesIO()
+        buf.length = buf.write(raw)
+        buf.seek(0)
 
+        if self.save is not None:
+            # concatenate buffers
             buf.seek(0)
+            self.save.write(buf.read())
+            self.save.length += buf.length
+            # discard save
+            buf = self.save
+            self.save = None
 
-            if save is not None:
-                # concatenate buffers
+        offset = 0
+        while offset < buf.length:
+            buf.seek(offset)
+            (length,
+             mtype,
+             flags) = struct.unpack('IHH', buf.read(8))
+
+            if offset + length > buf.length:
+                # create save buffer
+                buf.seek(offset)
+                self.save = io.BytesIO()
+                self.save.length = self.save.write(buf.read())
+                # truncate the buffer
+                buf.truncate(offset)
+                break
+
+            buf.seek(offset)
+            data = io.BytesIO()
+            data.write(buf.read(length))
+            data.length = length
+            data.seek(0)
+
+            # data traffic
+            envelope = envmsg(data)
+            envelope.decode()
+            nonce = envelope['header']['sequence_number']
+            try:
+                buf = io.BytesIO()
+                buf.length = buf.write(envelope.
+                                       get_attr('IPR_ATTR_CDATA'))
                 buf.seek(0)
-                save.write(buf.read())
-                save.length += buf.length
-                # discard save
-                buf = save
-                save = None
+                if ((flags & NLT_CONTROL) and
+                        (flags & NLT_RESPONSE)):
+                    msg = mgmtmsg(buf)
+                    msg.decode()
+                    self.listeners[nonce].put_nowait(msg)
+                else:
+                    self.parse(envelope, buf)
+            except AttributeError:
+                # now silently drop bad packet
+                pass
 
-            offset = 0
-            while offset < buf.length:
-                buf.seek(offset)
-                (length,
-                 mtype,
-                 flags) = struct.unpack('IHH', buf.read(8))
+            offset += length
 
-                if offset + length > buf.length:
-                    # create save buffer
-                    buf.seek(offset)
-                    save = io.BytesIO()
-                    save.length = save.write(buf.read())
-                    # truncate the buffer
-                    buf.truncate(offset)
-                    break
+    @debug
+    def parse(self, envelope, data):
 
-                buf.seek(offset)
-                data = io.BytesIO()
-                data.write(buf.read(length))
-                data.length = length
-                data.seek(0)
+        if self.marshal is None:
+            nonce = envelope['header']['sequence_number']
+            if envelope['header']['flags'] & NLT_EXCEPTION:
+                error = RuntimeError(data.getvalue())
+                msgs = [{'header': {'sequence_number': nonce,
+                                    'type': 0,
+                                    'flags': 0,
+                                    'error': error},
+                         'data': None}]
+            else:
+                msgs = [{'header': {'sequence_number': nonce,
+                                    'type': 0,
+                                    'flags': 0,
+                                    'error': None},
+                         'data': data.getvalue()}]
+        else:
+            msgs = self.marshal.parse(data)
 
-                # data traffic
-                envelope = envmsg(data)
-                envelope.decode()
-                nonce = envelope['header']['sequence_number']
-                try:
-                    buf = io.BytesIO()
-                    buf.length = buf.write(envelope.
-                                           get_attr('IPR_ATTR_CDATA'))
-                    buf.seek(0)
-                    if ((flags & NLT_CONTROL) and
-                            (flags & NLT_RESPONSE)):
-                        msg = mgmtmsg(buf)
-                        msg.decode()
-                        self.listeners[nonce].put_nowait(msg)
-                    elif self.marshal is None:
-                        msg = envelope.get_attr('IPR_ATTR_CDATA')
-                        self.listeners[0].put_nowait(msg)
-                    else:
-                        self.parse(buf)
-                except AttributeError:
-                    # now silently drop bad packet
-                    pass
-
-                offset += length
-
-    def parse(self, data):
-
-        for msg in self.marshal.parse(data):
-            key = msg['header']['sequence_number']
+        for msg in msgs:
+            try:
+                key = msg['header']['sequence_number']
+            except (TypeError, KeyError):
+                key = 0
 
             # 8<--------------------------------------------------------------
             # message filtering
             # right now it is simply iterating callback list
+            # .. _ioc-callbacks:
+            skip = False
+
             for cr in self.callbacks:
-                if cr[0](msg):
-                    cr[1](msg, *cr[2])
+                if cr[0](envelope, msg):
+                    if cr[1](envelope, msg, *cr[2]) is not None:
+                        skip = True
+
+            if skip:
+                continue
 
             # 8<--------------------------------------------------------------
             if key not in self.listeners:
                 key = 0
 
-            if self._mirror and (key != 0) and (msg.raw is not None):
+            if self._mirror and (key != 0):
                 # On Python 2.6 it can fail due to class fabrics
                 # in nlmsg definitions, so parse it again. It should
                 # not be much slower than copy.deepcopy()
-                try:
+                if getattr(msg, 'raw', None) is not None:
                     raw = io.BytesIO()
                     raw.length = raw.write(msg.raw)
-                    self.listeners[0].put_nowait(self.marshal.parse(raw)[0])
-                except Queue.Full:
-                    # FIXME: log this
-                    pass
+                    new = self.marshal.parse(raw)[0]
+                else:
+                    new = copy.deepcopy(msg)
+                self.listeners[0].put_nowait(new)
 
             if key in self.listeners:
                 try:
@@ -219,7 +218,11 @@ class IOCore(threading.Thread):
         msg = mgmtmsg(io.BytesIO())
         msg['cmd'] = cmd
         msg['attrs'] = attrs
-        rsp = self.nlm_request(msg, NLMSG_CONTROL, 0, NLT_CONTROL, addr)[0]
+        msg['header']['type'] = NLMSG_CONTROL
+        msg.encode()
+        rsp = self.request(msg.buf.getvalue(),
+                           env_flags=NLT_CONTROL,
+                           addr=addr)[0]
         assert rsp['cmd'] == IPRCMD_ACK
         if expect is not None:
             if type(expect) not in (list, tuple):
@@ -238,6 +241,7 @@ class IOCore(threading.Thread):
                             addr=addr)
 
     def discover(self, url, addr=None):
+        # .. _ioc-discover:
         return self.command(IPRCMD_DISCOVER,
                             [['IPR_ATTR_HOST', url]],
                             expect='IPR_ATTR_ADDR',
@@ -294,17 +298,14 @@ class IOCore(threading.Thread):
             except Queue.Empty as e:
                 if addr == self.default_broker:
                     raise e
-        self.iothread.stop()
+        self.iobroker.shutdown()
+        self.ioloop.shutdown()
+        self.ioloop.join()
 
-        self._stop_event.set()
         self._brs.send(struct.pack('I', 4))
         self._brs.close()
-        self.join()
         self.bridge.close()
 
-        self.buffers.put("")
-        self._feed_thread.join()
-
     def mirror(self, operate=True):
         '''
         Turn message mirroring on/off. When it is 'on', all
@@ -337,7 +338,8 @@ class IOCore(threading.Thread):
             self.cid = None
             del self.listeners[0]
 
-    def register_callback(self, callback, predicate=lambda x: True, args=None):
+    def register_callback(self, callback,
+                          predicate=lambda e, x: True, args=None):
         '''
         Register a callback to run on a message arrival.
 
@@ -350,7 +352,7 @@ class IOCore(threading.Thread):
         Simplest example, assume ipr is the IPRoute() instance::
 
             # create a simplest callback that will print messages
-            def cb(msg):
+            def cb(env, msg):
                 print(msg)
 
             # register callback for any message:
@@ -359,12 +361,12 @@ class IOCore(threading.Thread):
         More complex example, with filtering::
 
             # Set object's attribute after the message key
-            def cb(msg, obj):
+            def cb(env, msg, obj):
                 obj.some_attr = msg["some key"]
 
             # Register the callback only for the loopback device, index 1:
             ipr.register_callback(cb,
-                                  lambda x: x.get('index', None) == 1,
+                                  lambda e, x: x.get('index', None) == 1,
                                   (self, ))
 
         Please note: you do **not** need to register the default 0 queue
@@ -386,23 +388,8 @@ class IOCore(threading.Thread):
                 self.callbacks.pop(cb.index(cr))
                 return
 
-    def _remove_queue(self, key):
-        '''
-        Flush the queue to the default one and remove it
-        '''
-        queue = self.listeners[key]
-        # only not the default queue
-        if key != 0:
-            # delete the queue
-            del self.listeners[key]
-            # get remaining messages from the queue and
-            # re-route them to queue 0 or drop
-            while not queue.empty():
-                msg = queue.get()
-                if 0 in self.listeners:
-                    self.listeners[0].put(msg)
-
-    def get(self, key=0, raw=False, timeout=None):
+    @debug
+    def get(self, key=0, raw=False, timeout=None, terminate=None):
         '''
         Get a message from a queue
 
@@ -410,49 +397,68 @@ class IOCore(threading.Thread):
         '''
         queue = self.listeners[key]
         result = []
-        timeout = timeout or self._timeout
+        e = None
+        timeout = (timeout or self._timeout) if (key != 0) else 0xffff
         while True:
             # timeout should also be set to catch ctrl-c
             # Bug-Url: http://bugs.python.org/issue1360
             try:
                 msg = queue.get(block=True, timeout=timeout)
-            except Queue.Empty as e:
-                if key == 0 or hasattr(queue, 'persist'):
+            except Queue.Empty as x:
+                e = x
+                if key == 0:
                     continue
-                self._remove_queue(key)
-                raise e
+                else:
+                    break
+
+            if (terminate is not None) and terminate(msg):
+                break
+
+            # exceptions
+            if msg['header'].get('error', None) is not None:
+                e = msg['header']['error']
+
+            # RPC
             if self.marshal is None:
-                return [msg]
-            # terminator for persistent queues
-            if msg is None:
-                self._remove_queue(key)
-                raise Queue.Empty()
-            if (msg['header'].get('error', None) is not None) and\
-                    (not raw):
-                self._remove_queue(key)
-                raise msg['header']['error']
-            if (msg['header']['type'] != NLMSG_DONE) or raw:
-                result.append(msg)
-            if (msg['header']['type'] == NLMSG_DONE) or \
-               (not msg['header']['flags'] & NLM_F_MULTI):
+                data = msg.get('data', msg)
+            else:
+                data = msg
+
+            # Netlink
+            if (msg['header']['type'] != NLMSG_DONE):
+                result.append(data)
+
+            # break the loop if any
+            if (key == 0) or (e is not None):
                 break
-            if raw:
+
+            # wait for NLMSG_DONE if NLM_F_MULTI
+            if (terminate is None) and (
+                    (msg['header']['type'] == NLMSG_DONE) or
+                    (not msg['header']['flags'] & NLM_F_MULTI)):
                 break
-        if not hasattr(queue, 'persist'):
-            self._remove_queue(key)
+
+        if key != 0:
+            # delete the queue
+            del self.listeners[key]
+            self.nonce.free(key)
+            # get remaining messages from the queue and
+            # re-route them to queue 0 or drop
+            while not queue.empty():
+                msg = queue.get()
+                if 0 in self.listeners:
+                    self.listeners[0].put(msg)
+
+        if e is not None:
+            raise e
+
         return result
 
-    def nonce(self):
-        with self._nonce_lock:
-            if self._nonce == 0xffffffff:
-                self._nonce = 1
-            else:
-                self._nonce += 1
-            return self._nonce
-
+    @debug
     def push(self, host, msg,
              env_flags=None,
-             nonce=0):
+             nonce=0,
+             cname=None):
         addr, port = host
         envelope = envmsg()
         envelope['header']['sequence_number'] = nonce
@@ -465,6 +471,8 @@ class IOCore(threading.Thread):
         envelope['dport'] = port
         envelope['ttl'] = 16
         envelope['attrs'] = [['IPR_ATTR_CDATA', msg]]
+        if cname is not None:
+            envelope['attrs'].append(['IPR_ATTR_CNAME', cname])
         envelope.encode()
         self.bridge.send(envelope.buf.getvalue())
 
@@ -472,47 +480,15 @@ class IOCore(threading.Thread):
                 env_flags=0,
                 addr=None,
                 port=None,
-                response_timeout=None):
+                nonce=None,
+                cname=None,
+                response_timeout=None,
+                terminate=None):
+        nonce = nonce or self.nonce.alloc()
         port = port or self.default_dport
         addr = addr or self.default_broker
-        self.nlm_push(msg, env_flags, addr, port)
-        return self.get(0, timeout=response_timeout)
-
-    def nlm_push(self, msg,
-                 msg_type=None,
-                 msg_flags=None,
-                 env_flags=None,
-                 addr=0,
-                 port=0,
-                 nonce=0):
-        msg['header']['sequence_number'] = nonce
-        msg['header']['pid'] = os.getpid()
-        if msg_type is not None:
-            msg['header']['type'] = msg_type
-        if msg_flags is not None:
-            msg['header']['flags'] = msg_flags
-        msg.encode()
-        self.push((addr, port), msg.buf.getvalue(), env_flags, nonce)
-
-    def nlm_request(self, msg, msg_type,
-                    msg_flags=NLM_F_DUMP | NLM_F_REQUEST,
-                    env_flags=0,
-                    addr=None,
-                    port=None,
-                    response_timeout=None):
-        '''
-        Send netlink request, filling common message
-        fields, and wait for response.
-        '''
-        port = port or self.default_dport
-        addr = addr or self.default_peer
-        nonce = self.nonce()
         self.listeners[nonce] = Queue.Queue(maxsize=_QUEUE_MAXSIZE)
-        self.nlm_push(msg, msg_type, msg_flags, env_flags, addr, port, nonce)
-        result = self.get(nonce, timeout=response_timeout)
-        for msg in result:
-            # reset message buffer, make it ready for encoding back
-            msg.reset()
-            if not self.debug:
-                del msg['header']
-        return result
+        self.push((addr, port), msg, env_flags, nonce, cname)
+        return self.get(nonce,
+                        timeout=response_timeout,
+                        terminate=terminate)
